// ----
// ---- file   : aon_fm_v1.cpp
// ---- author : bsp
// ---- legal  : Distributed under terms of the MIT license (https://opensource.org/licenses/MIT)
// ----
// ----          Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
// ----          associated documentation files (the "Software"), to deal in the Software without restriction, including 
// ----          without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
// ----          copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to 
// ----          the following conditions:
// ----  
// ----          The above copyright notice and this permission notice shall be included in all copies or substantial 
// ----          portions of the Software.
// ----
// ----          THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
// ----          NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// ----          IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// ----          WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// ----          SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// ----
// ---- info   : auto-generated by "Cycle"
// ----           $ g++ -Wall -Wno-unused-function -Wno-unused-variable -I../../tksampler -c aon_fm_v1.cpp -o aon_fm_v1.o
// ---- created: 15Dec2023 17:07:26
// ----
// ----
// ----

#include <plugin.h>

#include <stdlib.h>
#include <math.h>
#include <string.h>


#define OVERSAMPLE_FACTOR  4.0f

#define PARAM_CARRIER_WAVE       0
#define PARAM_CARRIER_FREQ       1
#define PARAM_MOD_WAVE           2
#define PARAM_MOD_FREQ           3
#define PARAM_MOD_AMP            4
#define PARAM_MOD_KBD            5
#define PARAM_CARRIER_KBD        6
#define PARAM_MOD_FLT            7
#define NUM_PARAMS               8
static const char *loc_param_names[NUM_PARAMS] = {
   "carrier_wave",            // 0: CARRIER_WAVE
   "carrier_freq",            // 1: CARRIER_FREQ
   "mod_wave",                // 2: MOD_WAVE
   "mod_freq",                // 3: MOD_FREQ
   "mod_amp",                 // 4: MOD_AMP
   "mod_kbd",                 // 5: MOD_KBD
   "carrier_kbd",             // 6: CARRIER_KBD
   "mod_flt",                 // 7: MOD_FLT

};
static float loc_param_resets[NUM_PARAMS] = {
   0.0f,                      // 0: CARRIER_WAVE
   0.5f,                      // 1: CARRIER_FREQ
   0.0f,                      // 2: MOD_WAVE
   0.5f,                      // 3: MOD_FREQ
   0.5f,                      // 4: MOD_AMP
   0.625f,                    // 5: MOD_KBD
   0.625f,                    // 6: CARRIER_KBD
   0.8f,                      // 7: MOD_FLT

};

#define MOD_CARRIER_WAVE         0
#define MOD_CARRIER_FREQ         1
#define MOD_MOD_WAVE             2
#define MOD_MOD_FREQ             3
#define MOD_MOD_AMP              4
#define MOD_MOD_FLT              5
#define NUM_MODS                 6
static const char *loc_mod_names[NUM_MODS] = {
   "carrier_wave",         // 0: CARRIER_WAVE
   "carrier_freq",         // 1: CARRIER_FREQ
   "mod_wave",             // 2: MOD_WAVE
   "mod_freq",             // 3: MOD_FREQ
   "mod_amp",              // 4: MOD_AMP
   "mod_flt",              // 5: MOD_FLT

};

typedef struct aon_fm_v1_info_s {
   st_plugin_info_t base;
} aon_fm_v1_info_t;

typedef struct aon_fm_v1_shared_s {
   st_plugin_shared_t base;
   float params[NUM_PARAMS];

} aon_fm_v1_shared_t;

typedef struct aon_fm_v1_voice_s {
   st_plugin_voice_t base;
   float             sample_rate;
   float             note_speed_fixed;
   float             note_speed_cur;
   float             note_speed_inc;
   float             note_cur;
   float             note_inc;
   float             mods[NUM_MODS];
   float mod_carrier_wave_cur;
   float mod_carrier_wave_inc;
   float mod_carrier_freq_cur;
   float mod_carrier_freq_inc;
   float mod_mod_wave_cur;
   float mod_mod_wave_inc;
   float mod_mod_freq_cur;
   float mod_mod_freq_inc;
   float mod_mod_amp_cur;
   float mod_mod_amp_inc;
   float mod_mod_flt_cur;
   float mod_mod_flt_inc;

   float tmp1_kbd;
   float tmp2_off;
   float tmp3_scl;
   float tmp4_kbd;
   float tmp5_off;
   float tmp6_scl;
   float tmp7_kbd;
   float tmp8_off;
   float tmp9_scl;
   int tmp10_zlp_pos;
   float tmp11_zlp_sub_off_f;
   float tmp13_zlp_smp_next;
   float tmp14_rate;
   float tmp15_seq;
   float tmp16_seq;
   int tmp17_zlp_pos;
   float tmp18_zlp_sub_off_f;
   float tmp20_zlp_smp_next;
   float tmp21_rate;
   float tmp22_seq;
   float tmp23_seq;
   float tmp24_svf_lp;
   float tmp25_svf_hp;
   float tmp26_svf_bp;
   float tmp27_freq;
   float tmp28_seq;
   float tmp29_seq;
   float var_x;
   float var_v_carrier_kbd;
   float var_v_mod_kbd;
   float var_v_mod_flt;

} aon_fm_v1_voice_t;

#define loop(X)  for(unsigned int i = 0u; i < (X); i++)
#define clamp(a,b,c) (((a)<(b))?(b):(((a)>(c))?(c):(a)))

static inline float mathLerpf(float _a, float _b, float _t) { return _a + (_b - _a) * _t; }
static inline float mathClampf(float a, float b, float c) { return (((a)<(b))?(b):(((a)>(c))?(c):(a))); }
static inline float mathMinf(float a, float b) { return (a<b)?a:b; }
static inline float mathMaxf(float a, float b) { return (a>b)?a:b; }
static inline float mathAbsMaxf(float _x, float _y) { return ( ( (_x<0.0f)?-_x:_x)>((_y<0.0f)?-_y:_y)?_x:_y ); }
static inline float mathAbsMinf(float _x, float _y) { return ( ((_x<0.0f)?-_x:_x)<((_y<0.0f)?-_y:_y)?_x:_y ); }
static inline float frac(float _x) { return _x - ((int)_x); }

static inline float winLinear(const float *_s, float _index) {
   int idx = (int)_index;
   float r = _index - (float)idx;
   return mathLerpf(_s[idx], _s[idx+1], r);
}

static float ffrac_s(float _f) { int i; if(_f >= 0.0f) { i = (int)_f; return _f - (float)i; } else { i = (int)-_f; return 1.0f - (-_f - (float)i); } }

static float loc_bipolar_to_scale(const float _t, const float _div, const float _mul) {
   // t (-1..1) => /_div .. *_mul
   
   float s;

   if(_t < 0.0f)
   {
      s = (1.0f / _div);
      s = 1.0f + (s - 1.0f) * -_t;
      if(s < 0.0f)
         s = 0.0f;
   }
   else
   {
      s = _mul;
      s = 1.0f + (s - 1.0f) * _t;
   }
   
   return s;
}

static const float loc_smp_aon_fm_oscwaves[1536] = {
   0.0233459f, 0.0704041f, 0.117462f, 0.16452f, 0.211578f, 0.258636f, 0.305695f, 0.352753f
   , 0.399811f, 0.446869f, 0.486115f, 0.533142f, 0.572357f, 0.611572f, 0.650787f, 0.68219f
   , 0.721405f, 0.752777f, 0.784149f, 0.815521f, 0.83905f, 0.86261f, 0.886139f, 0.909637f
   , 0.925354f, 0.94104f, 0.956726f, 0.9646f, 0.980255f, 0.988098f, 0.995941f, 0.995972f
   , 0.995972f, 0.995972f, 0.988159f, 0.980316f, 0.972473f, 0.9646f, 0.948944f, 0.933258f
   , 0.909729f, 0.88623f, 0.862701f, 0.839142f, 0.8078f, 0.776428f, 0.745056f, 0.713684f
   , 0.6745f, 0.635284f, 0.603882f, 0.556854f, 0.517639f, 0.478394f, 0.431366f, 0.384308f
   , 0.33725f, 0.298035f, 0.250977f, 0.203918f, 0.15686f, 0.109802f, 0.0627441f, 0.015686f
   , -0.0233459f, -0.0704041f, -0.117462f, -0.16452f, -0.211578f, -0.258636f, -0.305695f, -0.352753f
   , -0.399811f, -0.446869f, -0.486084f, -0.533142f, -0.572357f, -0.611572f, -0.650787f, -0.68219f
   , -0.721375f, -0.752777f, -0.784149f, -0.807709f, -0.83905f, -0.86261f, -0.886139f, -0.909637f
   , -0.925354f, -0.94104f, -0.956726f, -0.9646f, -0.980255f, -0.988098f, -0.988129f, -0.988129f
   , -0.988129f, -0.988129f, -0.980316f, -0.972473f, -0.96463f, -0.956757f, -0.941101f, -0.925415f
   , -0.901917f, -0.8862f, -0.862671f, -0.831329f, -0.8078f, -0.776428f, -0.745056f, -0.713684f
   , -0.6745f, -0.643097f, -0.603882f, -0.564667f, -0.525452f, -0.478424f, -0.439178f, -0.399994f
   , -0.352936f, -0.305878f, -0.25882f, -0.211761f, -0.164703f, -0.117645f, -0.0705872f, -0.015686f
   , 0.0155334f, 0.0469055f, 0.0782776f, 0.10965f, 0.141022f, 0.172394f, 0.203766f, 0.235138f
   , 0.26651f, 0.297882f, 0.329254f, 0.360626f, 0.391998f, 0.42337f, 0.454742f, 0.486115f
   , 0.517487f, 0.548859f, 0.580231f, 0.611603f, 0.642975f, 0.674347f, 0.705719f, 0.737091f
   , 0.768463f, 0.799835f, 0.831207f, 0.862579f, 0.893951f, 0.925323f, 0.956696f, 0.988068f
   , 0.980347f, 0.948975f, 0.917603f, 0.88623f, 0.854858f, 0.823486f, 0.792114f, 0.760742f
   , 0.72937f, 0.697998f, 0.666626f, 0.635254f, 0.603882f, 0.57251f, 0.541138f, 0.509766f
   , 0.478394f, 0.447021f, 0.415649f, 0.384277f, 0.352905f, 0.321533f, 0.290161f, 0.258789f
   , 0.227417f, 0.196045f, 0.164673f, 0.133301f, 0.101929f, 0.0705566f, 0.0391846f, 0.0078125f
   , -0.0155334f, -0.0469055f, -0.0782776f, -0.10965f, -0.141022f, -0.172394f, -0.203766f, -0.235138f
   , -0.26651f, -0.297882f, -0.329254f, -0.360626f, -0.391998f, -0.42337f, -0.454742f, -0.486115f
   , -0.517487f, -0.548859f, -0.580231f, -0.611603f, -0.642975f, -0.674347f, -0.705719f, -0.737091f
   , -0.768463f, -0.799835f, -0.831207f, -0.862579f, -0.893951f, -0.925323f, -0.956696f, -0.99588f
   , -0.980316f, -0.948975f, -0.917603f, -0.88623f, -0.854858f, -0.823486f, -0.792114f, -0.760742f
   , -0.72937f, -0.697998f, -0.666626f, -0.635254f, -0.603882f, -0.57251f, -0.541138f, -0.509766f
   , -0.478394f, -0.447021f, -0.415649f, -0.384277f, -0.352905f, -0.321533f, -0.290161f, -0.258789f
   , -0.227417f, -0.196045f, -0.164673f, -0.133301f, -0.101929f, -0.0705566f, -0.0391846f, -0.0078125f
   , -0.995972f, -0.995972f, -0.988159f, -0.988129f, -0.980286f, -0.972443f, -0.96463f, -0.956787f
   , -0.948944f, -0.948914f, -0.941071f, -0.933228f, -0.925385f, -0.917542f, -0.909698f, -0.894043f
   , -0.8862f, -0.878357f, -0.870514f, -0.86264f, -0.846985f, -0.839142f, -0.831299f, -0.815613f
   , -0.80777f, -0.799896f, -0.784241f, -0.776367f, -0.760712f, -0.745026f, -0.737152f, -0.721497f
   , -0.705811f, -0.690125f, -0.674438f, -0.658752f, -0.643066f, -0.62738f, -0.611694f, -0.596008f
   , -0.572479f, -0.556793f, -0.533295f, -0.517578f, -0.49408f, -0.470551f, -0.454834f, -0.431305f
   , -0.407776f, -0.384277f, -0.360718f, -0.337189f, -0.31366f, -0.282318f, -0.258789f, -0.235229f
   , -0.203888f, -0.180359f, -0.148987f, -0.125458f, -0.0940857f, -0.0705261f, -0.0391846f, -0.0078125f
   , 0.00772095f, 0.0390625f, 0.0704346f, 0.0939636f, 0.125336f, 0.148895f, 0.180237f, 0.203796f
   , 0.235138f, 0.258667f, 0.282227f, 0.313568f, 0.337097f, 0.360626f, 0.384155f, 0.407715f
   , 0.431213f, 0.454742f, 0.470459f, 0.493988f, 0.517517f, 0.533203f, 0.556732f, 0.572418f
   , 0.595917f, 0.611633f, 0.627319f, 0.643005f, 0.658691f, 0.674377f, 0.690063f, 0.70575f
   , 0.721436f, 0.737122f, 0.744965f, 0.760651f, 0.776337f, 0.78418f, 0.799866f, 0.807709f
   , 0.815582f, 0.831238f, 0.839111f, 0.846954f, 0.86261f, 0.870453f, 0.878326f, 0.886169f
   , 0.894012f, 0.909668f, 0.917511f, 0.925354f, 0.933197f, 0.94104f, 0.948883f, 0.948914f
   , 0.956757f, 0.9646f, 0.972443f, 0.980255f, 0.988098f, 0.988129f, 0.995972f, 0.995972f
   , 0.972534f, 0.909821f, 0.847076f, 0.784332f, 0.729431f, 0.6745f, 0.611786f, 0.556885f
   , 0.501984f, 0.447083f, 0.392181f, 0.345093f, 0.290222f, 0.243134f, 0.196075f, 0.149017f
   , 0.101959f, 0.0549011f, 0.00784302f, -0.031189f, -0.0704041f, -0.117462f, -0.156677f, -0.195892f
   , -0.235107f, -0.274323f, -0.313538f, -0.34494f, -0.384125f, -0.415527f, -0.446899f, -0.478271f
   , -0.509644f, -0.541016f, -0.572388f, -0.60376f, -0.627289f, -0.658661f, -0.68219f, -0.705719f
   , -0.729248f, -0.752777f, -0.776306f, -0.792023f, -0.815521f, -0.831238f, -0.846924f, -0.86261f
   , -0.878296f, -0.893982f, -0.909668f, -0.925354f, -0.933197f, -0.948883f, -0.956726f, -0.964569f
   , -0.972412f, -0.980255f, -0.988098f, -0.988129f, -0.995941f, -0.995972f, -0.995972f, -0.995972f
   , -0.995972f, -0.995972f, -0.995972f, -0.995972f, -0.988159f, -0.988129f, -0.980286f, -0.972443f
   , -0.96463f, -0.956757f, -0.948914f, -0.933258f, -0.925415f, -0.909729f, -0.894043f, -0.886169f
   , -0.870514f, -0.846985f, -0.831299f, -0.815613f, -0.792114f, -0.776398f, -0.752869f, -0.72934f
   , -0.705841f, -0.682281f, -0.658752f, -0.627411f, -0.603882f, -0.57251f, -0.541138f, -0.517578f
   , -0.486237f, -0.447052f, -0.415649f, -0.384277f, -0.345093f, -0.31369f, -0.274475f, -0.235291f
   , -0.196075f, -0.15686f, -0.117615f, -0.0705872f, -0.0313721f, 0.0154724f, 0.054718f, 0.0939331f
   , 0.140991f, 0.195862f, 0.24295f, 0.290009f, 0.344879f, 0.391968f, 0.446869f, 0.50174f
   , 0.556641f, 0.611542f, 0.666473f, 0.729187f, 0.784088f, 0.846832f, 0.909576f, 0.964478f
   , -0.995972f, -0.995972f, -0.980316f, -0.972473f, -0.948975f, -0.933258f, -0.909729f, -0.878387f
   , -0.847015f, -0.815643f, -0.784271f, -0.752899f, -0.713684f, -0.6745f, -0.635284f, -0.596069f
   , -0.564667f, -0.525452f, -0.486237f, -0.447021f, -0.407806f, -0.368591f, -0.329407f, -0.298004f
   , -0.25882f, -0.227417f, -0.196045f, -0.164673f, -0.133301f, -0.101929f, -0.0783997f, -0.0548401f
   , -0.0234985f, -0.00778198f, 0.00772095f, 0.0312195f, 0.046936f, 0.0626221f, 0.0783081f, 0.0861511f
   , 0.101837f, 0.10968f, 0.117523f, 0.117554f, 0.125397f, 0.125397f, 0.13324f, 0.13324f
   , 0.125427f, 0.125397f, 0.125397f, 0.117554f, 0.109741f, 0.101898f, 0.0940552f, 0.0862122f
   , 0.0783691f, 0.0704956f, 0.0548401f, 0.0469666f, 0.031311f, 0.0234375f, 0.00778198f, 0.0f
   , -0.00772095f, -0.023407f, -0.03125f, -0.046936f, -0.0548096f, -0.0704651f, -0.0783081f, -0.0861816f
   , -0.101837f, -0.10968f, -0.117523f, -0.125366f, -0.125397f, -0.13324f, -0.13324f, -0.141083f
   , -0.141083f, -0.141083f, -0.141083f, -0.13324f, -0.13324f, -0.125397f, -0.117554f, -0.109711f
   , -0.0940552f, -0.0861816f, -0.0705261f, -0.0548401f, -0.031311f, -0.015625f, 0.00769043f, 0.0233765f
   , 0.0469055f, 0.0704651f, 0.101807f, 0.133179f, 0.156708f, 0.18808f, 0.227295f, 0.258667f
   , 0.297852f, 0.329254f, 0.368469f, 0.407654f, 0.446869f, 0.486084f, 0.525299f, 0.564514f
   , 0.603729f, 0.642944f, 0.674347f, 0.713562f, 0.752747f, 0.784149f, 0.815521f, 0.854736f
   , 0.878265f, 0.909637f, 0.933167f, 0.956696f, 0.972412f, 0.988098f, 0.995941f, 0.995972f
   , -0.949097f, -0.85498f, -0.753021f, -0.658905f, -0.556976f, -0.46283f, -0.368713f, -0.274597f
   , -0.188324f, -0.0942078f, -0.00793457f, 0.0625f, 0.148773f, 0.227203f, 0.305634f, 0.376221f
   , 0.446838f, 0.509583f, 0.572327f, 0.635071f, 0.689972f, 0.737061f, 0.784119f, 0.831177f
   , 0.862579f, 0.901794f, 0.925323f, 0.948883f, 0.972412f, 0.988098f, 0.995941f, 0.995972f
   , 0.995972f, 0.995972f, 0.988129f, 0.972473f, 0.948975f, 0.925446f, 0.901886f, 0.862732f
   , 0.831329f, 0.784302f, 0.737244f, 0.690186f, 0.635284f, 0.572571f, 0.509827f, 0.447083f
   , 0.376526f, 0.305908f, 0.227509f, 0.149078f, 0.0628357f, -0.0154114f, -0.0938721f, -0.187958f
   , -0.274261f, -0.368347f, -0.462463f, -0.55658f, -0.658539f, -0.752655f, -0.854584f, -0.94873f
   , 0.902039f, 0.80011f, 0.705963f, 0.611847f, 0.509888f, 0.415771f, 0.321655f, 0.235352f
   , 0.141266f, 0.0549622f, -0.0310974f, -0.109528f, -0.187988f, -0.266418f, -0.337036f, -0.415436f
   , -0.47821f, -0.540955f, -0.603699f, -0.65863f, -0.713531f, -0.76059f, -0.807648f, -0.846893f
   , -0.886078f, -0.909668f, -0.94101f, -0.964539f, -0.980255f, -0.988129f, -0.995972f, -0.995972f
   , -0.995972f, -0.988159f, -0.980316f, -0.96463f, -0.941101f, -0.90976f, -0.88623f, -0.847015f
   , -0.807831f, -0.760773f, -0.713715f, -0.658844f, -0.603943f, -0.541199f, -0.478455f, -0.41571f
   , -0.337311f, -0.266724f, -0.188293f, -0.109863f, -0.0235901f, 0.0546265f, 0.1409f, 0.235016f
   , 0.321289f, 0.415405f, 0.509521f, 0.61145f, 0.705597f, 0.799713f, 0.901672f, 0.995789f
   , -0.00772095f, -0.023407f, -0.039093f, -0.0547791f, -0.0704651f, -0.0861511f, -0.101837f, -0.117523f
   , -0.133209f, -0.148895f, -0.164581f, -0.180267f, -0.195953f, -0.211639f, -0.227325f, -0.243011f
   , -0.258698f, -0.274384f, -0.29007f, -0.305756f, -0.321442f, -0.337128f, -0.352814f, -0.3685f
   , -0.384186f, -0.399872f, -0.415558f, -0.431244f, -0.44693f, -0.462616f, -0.478302f, -0.493988f
   , 0.509674f, 0.52536f, 0.541046f, 0.556732f, 0.572418f, 0.588104f, 0.60379f, 0.619476f
   , 0.635162f, 0.650848f, 0.666534f, 0.68222f, 0.697906f, 0.713593f, 0.729279f, 0.744965f
   , 0.760651f, 0.776337f, 0.792023f, 0.807709f, 0.823395f, 0.839081f, 0.854767f, 0.870453f
   , 0.886139f, 0.901825f, 0.917511f, 0.933197f, 0.948883f, 0.964569f, 0.980255f, 0.995941f
   , -0.988159f, -0.972473f, -0.956787f, -0.941101f, -0.925415f, -0.909729f, -0.894043f, -0.878357f
   , -0.862671f, -0.846985f, -0.831299f, -0.815613f, -0.799927f, -0.784241f, -0.768555f, -0.752869f
   , -0.737183f, -0.721497f, -0.705811f, -0.690125f, -0.674438f, -0.658752f, -0.643066f, -0.62738f
   , -0.611694f, -0.596008f, -0.580322f, -0.564636f, -0.54895f, -0.533264f, -0.517578f, -0.501892f
   , 0.486206f, 0.47052f, 0.454834f, 0.439148f, 0.423462f, 0.407776f, 0.39209f, 0.376404f
   , 0.360718f, 0.345032f, 0.329346f, 0.31366f, 0.297974f, 0.282288f, 0.266602f, 0.250916f
   , 0.235229f, 0.219543f, 0.203857f, 0.188171f, 0.172485f, 0.156799f, 0.141113f, 0.125427f
   , 0.109741f, 0.0940552f, 0.0783691f, 0.0626831f, 0.0469971f, 0.031311f, 0.015625f, 0.0f
   , -0.964722f, -0.901978f, -0.839233f, -0.776489f, -0.713745f, -0.651001f, -0.588257f, -0.525513f
   , -0.462769f, -0.400024f, -0.33728f, -0.274536f, -0.211792f, -0.149048f, -0.0863037f, -0.0235596f
   , 0.0311584f, 0.0939026f, 0.156647f, 0.219391f, 0.282135f, 0.344879f, 0.407623f, 0.470367f
   , 0.533112f, 0.595856f, 0.6586f, 0.721344f, 0.784088f, 0.846832f, 0.909576f, 0.972321f
   , 0.980347f, 0.948975f, 0.917603f, 0.88623f, 0.854858f, 0.823486f, 0.792114f, 0.760742f
   , 0.72937f, 0.697998f, 0.666626f, 0.635254f, 0.603882f, 0.57251f, 0.541138f, 0.509766f
   , 0.478394f, 0.447021f, 0.415649f, 0.384277f, 0.352905f, 0.321533f, 0.290161f, 0.258789f
   , 0.227417f, 0.196045f, 0.164673f, 0.133301f, 0.101929f, 0.0705566f, 0.0391846f, 0.0078125f
   , 0.0155334f, 0.0469055f, 0.0782776f, 0.10965f, 0.141022f, 0.172394f, 0.203766f, 0.235138f
   , 0.26651f, 0.297882f, 0.329254f, 0.360626f, 0.391998f, 0.42337f, 0.454742f, 0.486115f
   , 0.517487f, 0.548859f, 0.580231f, 0.611603f, 0.642975f, 0.674347f, 0.705719f, 0.737091f
   , 0.768463f, 0.799835f, 0.831207f, 0.862579f, 0.893951f, 0.925323f, 0.956696f, 0.988068f
   , 0.949036f, 0.886292f, 0.823547f, 0.760803f, 0.698059f, 0.635315f, 0.572571f, 0.509827f
   , 0.447083f, 0.384338f, 0.321594f, 0.25885f, 0.196106f, 0.133362f, 0.0706177f, 0.00787354f
   , -0.0468445f, -0.109589f, -0.172333f, -0.235077f, -0.297821f, -0.360565f, -0.423309f, -0.486053f
   , -0.548798f, -0.611542f, -0.674286f, -0.73703f, -0.799774f, -0.862518f, -0.925262f, -0.988007f
   , 0.368713f, 0.322388f, 0.291382f, 0.24469f, 0.197998f, 0.16687f, 0.135742f, 0.104553f
   , 0.0734253f, 0.0578613f, 0.0266724f, 0.0111084f, -0.00418091f, -0.0197449f, -0.0353088f, -0.0665588f
   , -0.0821228f, -0.0976868f, -0.113251f, -0.128876f, -0.14444f, -0.160065f, -0.17569f, -0.191315f
   , -0.20694f, -0.222565f, -0.23819f, -0.253815f, -0.285065f, -0.285065f, -0.316315f, -0.33194f
   , -0.347626f, -0.363251f, -0.378876f, -0.394562f, -0.410248f, -0.425873f, -0.441559f, -0.457245f
   , -0.47287f, -0.488556f, -0.504242f, -0.519928f, -0.535614f, -0.5513f, -0.566986f, -0.582672f
   , -0.598358f, -0.614044f, -0.629791f, -0.645477f, -0.661163f, -0.692474f, -0.708221f, -0.739532f
   , -0.770905f, -0.786591f, -0.817963f, -0.833649f, -0.86496f, -0.896332f, -0.943329f, -0.974701f
   , -0.990387f, -0.959015f, -0.927643f, -0.880646f, -0.849335f, -0.833588f, -0.802277f, -0.78653f
   , -0.755219f, -0.723846f, -0.692535f, -0.676849f, -0.661163f, -0.645416f, -0.62973f, -0.598419f
   , -0.598358f, -0.567047f, -0.551361f, -0.535675f, -0.519989f, -0.504303f, -0.488617f, -0.472931f
   , -0.47287f, -0.457184f, -0.441559f, -0.410248f, -0.394562f, -0.378937f, -0.378876f, -0.363251f
   , -0.347626f, -0.316315f, -0.30069f, -0.285065f, -0.26944f, -0.253815f, -0.23819f, -0.20694f
   , -0.191315f, -0.17569f, -0.17569f, -0.160065f, -0.144501f, -0.128876f, -0.0976868f, -0.0976868f
   , -0.0664978f, -0.0509338f, -0.0353699f, -0.0198059f, -0.00424194f, 0.0267334f, 0.0422974f, 0.0734253f
   , 0.0889893f, 0.120178f, 0.151306f, 0.182434f, 0.213562f, 0.260193f, 0.306885f, 0.353394f
   , -0.988159f, -0.972473f, -0.956787f, -0.941101f, -0.925415f, -0.909729f, -0.894043f, -0.878357f
   , -0.862671f, -0.846985f, -0.831299f, -0.815613f, -0.799927f, -0.784241f, -0.768555f, -0.752869f
   , -0.737183f, -0.721497f, -0.705811f, -0.690125f, -0.674438f, -0.658752f, -0.643066f, -0.62738f
   , -0.611694f, -0.596008f, -0.580322f, -0.564636f, -0.54895f, -0.533264f, -0.517578f, -0.501892f
   , -0.486206f, -0.47052f, -0.454834f, -0.439148f, -0.423462f, -0.407776f, -0.39209f, -0.376404f
   , -0.360718f, -0.345032f, -0.329346f, -0.31366f, -0.297974f, -0.282288f, -0.266602f, -0.250916f
   , -0.235229f, -0.219543f, -0.203857f, -0.188171f, -0.172485f, -0.156799f, -0.141113f, -0.125427f
   , -0.109741f, -0.0940552f, -0.0783691f, -0.0626831f, -0.0469971f, -0.031311f, -0.015625f, 0.0f
   , 0.00772095f, 0.023407f, 0.039093f, 0.0547791f, 0.0704651f, 0.0861511f, 0.101837f, 0.117523f
   , 0.133209f, 0.148895f, 0.164581f, 0.180267f, 0.195953f, 0.211639f, 0.227325f, 0.243011f
   , 0.258698f, 0.274384f, 0.29007f, 0.305756f, 0.321442f, 0.337128f, 0.352814f, 0.3685f
   , 0.384186f, 0.399872f, 0.415558f, 0.431244f, 0.44693f, 0.462616f, 0.478302f, 0.493988f
   , 0.509674f, 0.52536f, 0.541046f, 0.556732f, 0.572418f, 0.588104f, 0.60379f, 0.619476f
   , 0.635162f, 0.650848f, 0.666534f, 0.68222f, 0.697906f, 0.713593f, 0.729279f, 0.744965f
   , 0.760651f, 0.776337f, 0.792023f, 0.807709f, 0.823395f, 0.839081f, 0.854767f, 0.870453f
   , 0.886139f, 0.901825f, 0.917511f, 0.933197f, 0.948883f, 0.964569f, 0.980255f, 0.995941f
   , -0.769135f, -0.627594f, -0.384796f, -0.235382f, -0.133392f, -0.164581f, -0.180298f, -0.141144f
   , -0.0942078f, 0.0780029f, 0.227081f, 0.321259f, 0.45459f, 0.47052f, 0.431274f, 0.540924f
   , 0.556702f, 0.572449f, 0.60376f, 0.736816f, 0.776398f, 0.659088f, 0.517731f, 0.486267f
   , 0.415802f, 0.345123f, 0.368469f, 0.345093f, 0.313629f, 0.211945f, 0.24292f, 0.368286f
   , 0.329559f, 0.172791f, 0.117523f, 0.125427f, 0.180206f, 0.203735f, 0.172577f, 0.156769f
   , 0.203705f, 0.196045f, 0.250763f, 0.188263f, 0.102081f, 0.00787354f, -0.023407f, 0.0f
   , 0.117188f, 0.188049f, 0.243011f, 0.242981f, 0.235229f, 0.149139f, 0.180145f, 0.282104f
   , 0.384094f, 0.415405f, 0.493896f, 0.61145f, 0.713379f, 0.776306f, 0.768616f, 0.729279f
   , 0.807587f, 0.878296f, 0.807831f, 0.752899f, 0.760742f, 0.682373f, 0.658783f, 0.776062f
   , 0.901642f, 0.980225f, 0.956848f, 0.831543f, 0.706024f, 0.588318f, 0.509827f, 0.517548f
   , 0.587891f, 0.650757f, 0.729187f, 0.752777f, 0.596405f, 0.408112f, 0.274506f, 0.23526f
   , 0.289948f, 0.368347f, 0.493835f, 0.564545f, 0.470734f, 0.462646f, 0.44696f, 0.407867f
   , 0.360809f, 0.219818f, 0.141266f, 0.0705872f, -0.0780029f, -0.10968f, -0.117554f, -0.0628052f
   , 0.0310059f, -0.0154114f, 0.0938416f, 0.227142f, 0.243042f, 0.133514f, 0.0314026f, -0.0311279f
   , -0.023407f, 0.0545349f, 0.148773f, 0.289825f, 0.290161f, 0.17276f, 0.000213623f, -0.156494f
   , -0.195984f, -0.0942993f, 0.0232544f, 0.11731f, 0.195862f, 0.250824f, 0.164764f, 0.000274658f
   , 0.0389709f, 0.125214f, 0.203674f, 0.282104f, 0.368378f, 0.438965f, 0.517395f, 0.587982f
   , 0.650757f, 0.713501f, 0.768402f, 0.815491f, 0.862549f, 0.901764f, 0.933167f, 0.956726f
   , 0.980255f, 0.988129f, 0.995972f, 0.995972f, 0.995972f, 0.980316f, 0.96463f, 0.941101f
   , 0.90976f, 0.878387f, 0.839172f, 0.799957f, 0.75293f, 0.698059f, 0.65097f, 0.596069f
   , 0.541168f, 0.478455f, 0.423553f, 0.368622f, 0.305908f, 0.251007f, 0.196106f, 0.149017f
   , 0.0941162f, 0.0470581f, 0.0f, -0.031189f, -0.0704041f, -0.101807f, -0.125366f, -0.148895f
   , -0.172424f, -0.18811f, -0.195984f, -0.203827f, -0.203827f, -0.203827f, -0.196014f, -0.188171f
   , -0.172485f, -0.156799f, -0.133301f, -0.117584f, -0.0940552f, -0.0627136f, -0.0391846f, -0.0078125f
   , 0.00772095f, 0.0390625f, 0.0626221f, 0.0939636f, 0.117493f, 0.133209f, 0.156738f, 0.172424f
   , 0.18811f, 0.195984f, 0.203827f, 0.203827f, 0.203827f, 0.196014f, 0.188171f, 0.172485f
   , 0.148987f, 0.125458f, 0.101929f, 0.0705566f, 0.0313416f, -0.00765991f, -0.046875f, -0.0939331f
   , -0.148804f, -0.195892f, -0.250793f, -0.305695f, -0.368408f, -0.423309f, -0.478241f, -0.540955f
   , -0.595856f, -0.650757f, -0.697845f, -0.752747f, -0.799805f, -0.83902f, -0.878235f, -0.909637f
   , -0.94101f, -0.964539f, -0.980255f, -0.995941f, -0.995972f, -0.995972f, -0.988159f, -0.980316f
   , -0.956818f, -0.933289f, -0.901917f, -0.862701f, -0.815674f, -0.768616f, -0.713715f, -0.651001f
   , -0.588257f, -0.51767f, -0.43927f, -0.368652f, -0.28244f, -0.203979f, -0.125549f, -0.0392761f
};
static const float *loc_smp_aon_fm_oscwaves_zone0 = loc_smp_aon_fm_oscwaves + 0;  // ref 1/1, zone 1/14 #frames=1536
static void loc_prepare(st_plugin_voice_t *_voice);

void loc_prepare(st_plugin_voice_t *_voice) {
   ST_PLUGIN_VOICE_CAST(aon_fm_v1_voice_t);
   ST_PLUGIN_VOICE_SHARED_CAST(aon_fm_v1_shared_t);

   int tmp10_zlp_pos;
   int tmp17_zlp_pos;
   float out = 0.0f;
   // -------- lane "prepare" modIdx=0
   
   // -- mod="1" dstVar=out
   out = 1.0f;
   
   // -- mod="kbd" dstVar=out
   out *= 127.0f;
   
   // ---- mod="kbd" input "kbd" seq 1/1
   
   // -- mod="$p_carrier_kbd" dstVar=voice->tmp1_kbd
   voice->tmp1_kbd = shared->params[PARAM_CARRIER_KBD] * 8.0f - 4.0f;
   
   // ---- mod="kbd" input "off" seq 1/1
   
   // -- mod="0.0833333" dstVar=voice->tmp2_off
   voice->tmp2_off = 0.0833333f;
   voice->tmp3_scl = 4;
   out += (voice->note_cur + (voice->tmp2_off * 12.0f) - 64.0f) * voice->tmp1_kbd;
   #ifdef OVERSAMPLE_FACTOR
   out = ((440.0f/32.0f)*expf( ((out-9.0f)/12.0f)*logf(2.0f) )) / (voice->sample_rate * OVERSAMPLE_FACTOR);
   #else
   out = ((440.0f/32.0f)*expf( ((out-9.0f)/12.0f)*logf(2.0f) )) / voice->sample_rate;
   #endif
   out *= voice->tmp3_scl;
   
   // -- mod="sto v_carrier_kbd" dstVar=out
   voice->var_v_carrier_kbd = out;
   
   // -- mod="1" dstVar=out
   out = 1.0f;
   
   // -- mod="kbd" dstVar=out
   out *= 127.0f;
   
   // ---- mod="kbd" input "kbd" seq 1/1
   
   // -- mod="$p_mod_kbd" dstVar=voice->tmp4_kbd
   voice->tmp4_kbd = shared->params[PARAM_MOD_KBD] * 8.0f - 4.0f;
   
   // ---- mod="kbd" input "off" seq 1/1
   
   // -- mod="0.0833333" dstVar=voice->tmp5_off
   voice->tmp5_off = 0.0833333f;
   voice->tmp6_scl = 4;
   out += (voice->note_cur + (voice->tmp5_off * 12.0f) - 64.0f) * voice->tmp4_kbd;
   #ifdef OVERSAMPLE_FACTOR
   out = ((440.0f/32.0f)*expf( ((out-9.0f)/12.0f)*logf(2.0f) )) / (voice->sample_rate * OVERSAMPLE_FACTOR);
   #else
   out = ((440.0f/32.0f)*expf( ((out-9.0f)/12.0f)*logf(2.0f) )) / voice->sample_rate;
   #endif
   out *= voice->tmp6_scl;
   
   // -- mod="sto v_mod_kbd" dstVar=out
   voice->var_v_mod_kbd = out;
   
   // -- mod="$m_mod_flt" dstVar=out
   out = voice->mod_mod_flt_cur;
   
   // -- mod="kbd" dstVar=out
   out *= 127.0f;
   
   // ---- mod="kbd" input "kbd" seq 1/1
   
   // -- mod="1" dstVar=voice->tmp7_kbd
   voice->tmp7_kbd = 1.0f;
   
   // ---- mod="kbd" input "off" seq 1/1
   
   // -- mod="0" dstVar=voice->tmp8_off
   voice->tmp8_off = 0.0f;
   voice->tmp9_scl = 4;
   out += (voice->note_cur + (voice->tmp8_off * 12.0f) - 64.0f) * voice->tmp7_kbd;
   out = clamp(out, 0.0f, 127.0f);
   #ifdef OVERSAMPLE_FACTOR
   out = ((440.0f/32.0f)*expf( ((out-9.0f)/12.0f)*logf(2.0f) )) / (voice->sample_rate * OVERSAMPLE_FACTOR);
   #else
   out = ((440.0f/32.0f)*expf( ((out-9.0f)/12.0f)*logf(2.0f) )) / voice->sample_rate;
   #endif
   out *= voice->tmp9_scl;
   
   // -- mod="sto v_mod_flt" dstVar=out
   voice->var_v_mod_flt = out;
} /* end prepare */

static unsigned int loc_copy_chars(char *_d, const unsigned int _dSize, const char *_s) {
   unsigned int r = 0u;
   if(NULL != _d)
   {
      // Write min(dSize,len(s)) characters to 'd'
      char *d = _d;
      if(NULL != _s)
      {
         const char *s = _s;
         for(;;)
         {
            if(r < _dSize)
            {
               char c = s[r++];
               *d++ = c;
               if(0 == c)
                  break;
            }
            else
            {
               if(r > 0u)
                  d[-1] = 0;  // force ASCIIZ
               break;
            }
         }
      }
      else
      {
         if(_dSize > 0u)
         {
            _d[0] = 0;
            r = 1u;
         }
      }
   }
   else
   {
      // Return total length of source string
      if(NULL != _s)
      {
         const char *s = _s;
         for(;;)
         {
            char c = s[r++];
            if(0 == c)
               break;
         }
      }
   }
   return r;
}

static unsigned int loc_copy_floats(float              *_d,
                                    const unsigned int  _dSize,
                                    const float        *_s,
                                    const unsigned int  _sNum
                                    ) {
   unsigned int r = 0u;
   if(NULL != _d)
   {
      // Write min(dSize,sNum) characters to 'd'
      if(NULL != _s)
      {
         unsigned int num = (_sNum > _dSize) ? _dSize : _sNum;
         memcpy((void*)_d, (const void*)_s, num * sizeof(float));
         r = num;
      }
   }
   else
   {
      // Query total number of presets
      r = _sNum;
   }
   return r;
}

static const char *ST_PLUGIN_API loc_get_param_name(st_plugin_info_t *_info,
                                                    unsigned int      _paramIdx
                                                    ) {
   (void)_info;
   return loc_param_names[_paramIdx];
}

static const char *ST_PLUGIN_API loc_get_param_group_name(st_plugin_info_t *_info,
                                                          unsigned int      _paramGroupIdx
                                                          ) {
   (void)_info;
   const char *r = NULL;
   static const char *groupNames[4] = { "shape", "freq", "amp", "filter" };
   if(_paramGroupIdx < 4u)
      r = groupNames[_paramGroupIdx];
   return r;
}

static unsigned int ST_PLUGIN_API loc_get_param_group_idx(st_plugin_info_t *_info,
                                                          unsigned int      _paramIdx
                                                          ) {
   (void)_info;
   unsigned int r = ~0u;
   static unsigned int groupIndices[NUM_PARAMS] = { 0, 1, 0, 1, 2, ~0u, ~0u, 3 };
   r = groupIndices[_paramIdx];
   return r;
}

static float ST_PLUGIN_API loc_get_param_reset(st_plugin_info_t *_info,
                                               unsigned int      _paramIdx
                                               ) {
   (void)_info;
   return loc_param_resets[_paramIdx];
}

static float ST_PLUGIN_API loc_get_param_value(st_plugin_shared_t *_shared,
                                               unsigned int        _paramIdx
                                               ) {
   ST_PLUGIN_SHARED_CAST(aon_fm_v1_shared_t);
   return shared->params[_paramIdx];
}

static void ST_PLUGIN_API loc_set_param_value(st_plugin_shared_t *_shared,
                                              unsigned int        _paramIdx,
                                              float               _value
                                              ) {
   ST_PLUGIN_SHARED_CAST(aon_fm_v1_shared_t);
   shared->params[_paramIdx] = _value;
}

static unsigned int ST_PLUGIN_API loc_query_dynamic_param_preset_values(st_plugin_shared_t *_shared,
                                                                        const unsigned int  _paramIdx,
                                                                        float              *_retValues,
                                                                        const unsigned int  _retValuesSize
                                                                        ) {
   ST_PLUGIN_SHARED_CAST(aon_fm_v1_shared_t);
   unsigned int r = 0u;
   if(PARAM_CARRIER_WAVE == _paramIdx)
   {
      static const float presetValues_0[12] = { 0.0f, 0.090909f, 0.181818f, 0.272727f, 0.363636f, 0.454545f, 0.545454f, 0.636363f, 0.727272f, 0.818181f, 0.90909f, 1.0f };
      r = loc_copy_floats(_retValues, _retValuesSize, presetValues_0, 12);
   }
   if(PARAM_MOD_WAVE == _paramIdx)
   {
      static const float presetValues_2[12] = { 0.0f, 0.090909f, 0.181818f, 0.272727f, 0.363636f, 0.454545f, 0.545454f, 0.636363f, 0.727272f, 0.818181f, 0.90909f, 1.0f };
      r = loc_copy_floats(_retValues, _retValuesSize, presetValues_2, 12);
   }

   return r;
}

static unsigned int ST_PLUGIN_API loc_query_dynamic_param_preset_name(st_plugin_shared_t *_shared,
                                                                      const unsigned int  _paramIdx,
                                                                      const unsigned int  _presetIdx,
                                                                      char               *_retBuf,
                                                                      const unsigned int  _retBufSize
                                                                      ) {
   ST_PLUGIN_SHARED_CAST(aon_fm_v1_shared_t);
   unsigned int r = 0u;
   if(PARAM_CARRIER_WAVE == _paramIdx)
   {
      static const char *presetName_0[12] = { "Sinus", "Triangle", "Turnpoint", "Parabola", "Extremes", "Up&Down", "LineMix", "M-Shape", "Heart", "Sawtooth", "Sample", "DoubleSinus" };
      r = loc_copy_chars(_retBuf, _retBufSize, presetName_0[_presetIdx]);
   }
   if(PARAM_MOD_WAVE == _paramIdx)
   {
      static const char *presetName_2[12] = { "Sinus", "Triangle", "Turnpoint", "Parabola", "Extremes", "Up&Down", "LineMix", "M-Shape", "Heart", "Sawtooth", "Sample", "DoubleSinus" };
      r = loc_copy_chars(_retBuf, _retBufSize, presetName_2[_presetIdx]);
   }

   return r;
}

static const char *ST_PLUGIN_API loc_get_mod_name(st_plugin_info_t *_info,
                                                  unsigned int      _modIdx
                                                  ) {
   (void)_info;
   return loc_mod_names[_modIdx];
}

static void ST_PLUGIN_API loc_set_sample_rate(st_plugin_voice_t *_voice,
                                              float              _sampleRate
                                              ) {
   ST_PLUGIN_VOICE_CAST(aon_fm_v1_voice_t);
   voice->sample_rate = _sampleRate;
}

static void ST_PLUGIN_API loc_note_on(st_plugin_voice_t  *_voice,
                                      int                 _bGlide,
                                      unsigned char       _note,
                                      float               _vel
                                      ) {
   ST_PLUGIN_VOICE_CAST(aon_fm_v1_voice_t);
   ST_PLUGIN_VOICE_SHARED_CAST(aon_fm_v1_shared_t);
   (void)_bGlide;
   (void)_note;
   (void)_vel;
   if(!_bGlide)
   {
      memset((void*)voice->mods, 0, sizeof(voice->mods));
#ifdef OVERSAMPLE_FACTOR
      voice->note_speed_fixed = (0.5f / (voice->sample_rate * OVERSAMPLE_FACTOR));
#else
      voice->note_speed_fixed = (0.5f / voice->sample_rate);
#endif // OVERSAMPLE_FACTOR
   int tmp10_zlp_pos;
   int tmp17_zlp_pos;


      voice->tmp10_zlp_pos = 0;
      voice->tmp17_zlp_pos = 0;
      voice->tmp24_svf_lp = 0.0f;
      voice->tmp25_svf_hp = 0.0f;
      voice->tmp26_svf_bp = 0.0f;

      voice->var_x = 0.0f;
      voice->var_v_carrier_kbd = 0.0f;
      voice->var_v_mod_kbd = 0.0f;
      voice->var_v_mod_flt = 0.0f;


   }
}

static void ST_PLUGIN_API loc_set_mod_value(st_plugin_voice_t *_voice,
                                            unsigned int       _modIdx,
                                            float              _value,
                                            unsigned           _frameOffset
                                            ) {
   ST_PLUGIN_VOICE_CAST(aon_fm_v1_voice_t);
   (void)_frameOffset;
   voice->mods[_modIdx] = _value;
}

static void ST_PLUGIN_API loc_prepare_block(st_plugin_voice_t *_voice,
                                            unsigned int       _numFrames,
                                            float              _freqHz,
                                            float              _note,
                                            float              _vol,
                                            float              _pan
                                            ) {
   ST_PLUGIN_VOICE_CAST(aon_fm_v1_voice_t);
   ST_PLUGIN_VOICE_SHARED_CAST(aon_fm_v1_shared_t);
   (void)_note;
   (void)_vol;
   (void)_pan;

#ifdef OVERSAMPLE_FACTOR
   float noteSpeed = _freqHz / (voice->sample_rate * OVERSAMPLE_FACTOR);
#else
   float noteSpeed = _freqHz / voice->sample_rate;
#endif // OVERSAMPLE_FACTOR

   float modcarrier_wave = shared->params[PARAM_CARRIER_WAVE] * 11.0f               + voice->mods[MOD_CARRIER_WAVE ];
   float modcarrier_freq = shared->params[PARAM_CARRIER_FREQ] * 2.0f - 1.0f         + voice->mods[MOD_CARRIER_FREQ ];
   float modmod_wave     = shared->params[PARAM_MOD_WAVE    ] * 11.0f               + voice->mods[MOD_MOD_WAVE     ];
   float modmod_freq     = shared->params[PARAM_MOD_FREQ    ] * 2.0f - 1.0f         + voice->mods[MOD_MOD_FREQ     ];
   float modmod_amp      = shared->params[PARAM_MOD_AMP     ] * 32.0f - 16.0f       + voice->mods[MOD_MOD_AMP      ];
   float modmod_flt      = shared->params[PARAM_MOD_FLT     ]                       + voice->mods[MOD_MOD_FLT      ];

   if(_numFrames > 0u)
   {
      // lerp
      float recBlockSize = (1.0f / _numFrames);
      voice->note_speed_inc = (noteSpeed - voice->note_speed_cur) * recBlockSize;
      voice->note_inc       = (_note - voice->note_cur) * recBlockSize;
      voice->mod_carrier_wave_inc = (modcarrier_wave    - voice->mod_carrier_wave_cur  ) * recBlockSize;
      voice->mod_carrier_freq_inc = (modcarrier_freq    - voice->mod_carrier_freq_cur  ) * recBlockSize;
      voice->mod_mod_wave_inc     = (modmod_wave        - voice->mod_mod_wave_cur      ) * recBlockSize;
      voice->mod_mod_freq_inc     = (modmod_freq        - voice->mod_mod_freq_cur      ) * recBlockSize;
      voice->mod_mod_amp_inc      = (modmod_amp         - voice->mod_mod_amp_cur       ) * recBlockSize;
      voice->mod_mod_flt_inc      = (modmod_flt         - voice->mod_mod_flt_cur       ) * recBlockSize;

      loc_prepare(&voice->base);
   }
   else
   {
      // initial params/modulation (first block, not rendered)
      voice->note_speed_cur = noteSpeed;
      voice->note_speed_inc = 0.0f;
      voice->note_cur       = _note;
      voice->note_inc       = 0.0f;
      voice->mod_carrier_wave_cur = modcarrier_wave;
      voice->mod_carrier_wave_inc = 0.0f;
      voice->mod_carrier_freq_cur = modcarrier_freq;
      voice->mod_carrier_freq_inc = 0.0f;
      voice->mod_mod_wave_cur     = modmod_wave;
      voice->mod_mod_wave_inc     = 0.0f;
      voice->mod_mod_freq_cur     = modmod_freq;
      voice->mod_mod_freq_inc     = 0.0f;
      voice->mod_mod_amp_cur      = modmod_amp;
      voice->mod_mod_amp_inc      = 0.0f;
      voice->mod_mod_flt_cur      = modmod_flt;
      voice->mod_mod_flt_inc      = 0.0f;

      loc_prepare(&voice->base);
   }

   // printf("xxx note_cur=%f\n", voice->note_cur);
}

static void ST_PLUGIN_API loc_process_replace(st_plugin_voice_t  *_voice,
                                              int                 _bMonoIn,
                                              const float        *_samplesIn,
                                              float              *_samplesOut, 
                                              unsigned int        _numFrames
                                              ) {
   (void)_bMonoIn;
   (void)_samplesIn;

   ST_PLUGIN_VOICE_CAST(aon_fm_v1_voice_t);
   ST_PLUGIN_VOICE_SHARED_CAST(aon_fm_v1_shared_t);

   // Mono output (replicate left to right channel)
   unsigned int j = 0u;
   unsigned int jStep = _bMonoIn ? 1u : 2u;
   unsigned int k = 0u;
   for(unsigned int i = 0u; i < _numFrames; i++)
   {
      float out = _samplesIn[j];
      j += jStep;
#ifdef OVERSAMPLE_FACTOR
      float outOS = 0.0f;
#endif // OVERSAMPLE_FACTOR

#ifdef OVERSAMPLE_FACTOR
      for(unsigned int osi = 0u; osi < OVERSAMPLE_FACTOR; osi++)
#endif // OVERSAMPLE_FACTOR
      {
         float tmp_f;
         float tmp2_f;
         
         // ========
         // ======== lane "out" modIdx=0
         // ========
         
         // -- mod="zlp" dstVar=out
         
         // ---- mod="zlp" input "sub" seq 1/1
         
         // -- mod="$m_carrier_wave" dstVar=voice->tmp11_zlp_sub_off_f
         voice->tmp11_zlp_sub_off_f = voice->mod_carrier_wave_cur;
         int tmp12_zlp_sub_off;
         tmp12_zlp_sub_off = (int)(voice->tmp11_zlp_sub_off_f * 128);
         if((tmp12_zlp_sub_off + 128) > 1536)
            tmp12_zlp_sub_off = 1408;
         else if(tmp12_zlp_sub_off < 0)
            tmp12_zlp_sub_off = 0;
         voice->tmp13_zlp_smp_next = loc_smp_aon_fm_oscwaves_zone0[tmp12_zlp_sub_off+(((voice->tmp10_zlp_pos >> 16) + 1) & 127)];
         out = loc_smp_aon_fm_oscwaves_zone0[tmp12_zlp_sub_off+( (voice->tmp10_zlp_pos >> 16)      & 127)];
         out = out + (voice->tmp13_zlp_smp_next - out) * ((voice->tmp10_zlp_pos & 65535) * (1.0f / 65536.0f));
         
         // ---- mod="zlp" input "rate" seq 1/1
         
         // -- mod="$p_carrier_freq" dstVar=voice->tmp14_rate
         voice->tmp14_rate = shared->params[PARAM_CARRIER_FREQ] * 2.0f - 1.0f;
         
         // -- mod="$m_carrier_freq" dstVar=voice->tmp14_rate
         voice->tmp15_seq = voice->mod_carrier_freq_cur;
         voice->tmp14_rate += voice->tmp15_seq;
         
         // -- mod="bts" dstVar=voice->tmp14_rate
         voice->tmp14_rate = loc_bipolar_to_scale(voice->tmp14_rate, 16.0f, 16.0f);
         
         // -- mod="zlp" dstVar=voice->tmp14_rate
         
         // ---- mod="zlp" input "sub" seq 1/1
         
         // -- mod="$m_mod_wave" dstVar=voice->tmp18_zlp_sub_off_f
         voice->tmp18_zlp_sub_off_f = voice->mod_mod_wave_cur;
         int tmp19_zlp_sub_off;
         tmp19_zlp_sub_off = (int)(voice->tmp18_zlp_sub_off_f * 128);
         if((tmp19_zlp_sub_off + 128) > 1536)
            tmp19_zlp_sub_off = 1408;
         else if(tmp19_zlp_sub_off < 0)
            tmp19_zlp_sub_off = 0;
         voice->tmp20_zlp_smp_next = loc_smp_aon_fm_oscwaves_zone0[tmp19_zlp_sub_off+(((voice->tmp17_zlp_pos >> 16) + 1) & 127)];
         voice->tmp16_seq = loc_smp_aon_fm_oscwaves_zone0[tmp19_zlp_sub_off+( (voice->tmp17_zlp_pos >> 16)      & 127)];
         voice->tmp16_seq = voice->tmp16_seq + (voice->tmp20_zlp_smp_next - voice->tmp16_seq) * ((voice->tmp17_zlp_pos & 65535) * (1.0f / 65536.0f));
         
         // ---- mod="zlp" input "rate" seq 1/1
         
         // -- mod="$p_mod_freq" dstVar=voice->tmp21_rate
         voice->tmp21_rate = shared->params[PARAM_MOD_FREQ] * 2.0f - 1.0f;
         
         // -- mod="$m_mod_freq" dstVar=voice->tmp21_rate
         voice->tmp22_seq = voice->mod_mod_freq_cur;
         voice->tmp21_rate += voice->tmp22_seq;
         
         // -- mod="bts" dstVar=voice->tmp21_rate
         voice->tmp21_rate = loc_bipolar_to_scale(voice->tmp21_rate, 16.0f, 16.0f);
         
         // -- mod="$v_mod_kbd" dstVar=voice->tmp21_rate
         voice->tmp23_seq = voice->var_v_mod_kbd;
         voice->tmp21_rate *= voice->tmp23_seq;
         voice->tmp17_zlp_pos += int(voice->tmp21_rate * 65536);
         
         // -- mod="svf" dstVar=voice->tmp16_seq
         
         // ---- mod="svf" input "freq" seq 1/1
         
         // -- mod="$v_mod_flt" dstVar=voice->tmp27_freq
         voice->tmp27_freq = voice->var_v_mod_flt;
         voice->tmp24_svf_lp = voice->tmp24_svf_lp + (voice->tmp26_svf_bp * voice->tmp27_freq);
         voice->tmp25_svf_hp = voice->tmp16_seq - voice->tmp24_svf_lp - (voice->tmp26_svf_bp * 1.0f);
         voice->tmp26_svf_bp = voice->tmp26_svf_bp + (voice->tmp25_svf_hp * voice->tmp27_freq);
         voice->tmp16_seq = voice->tmp24_svf_lp;
         
         // -- mod="$m_mod_amp" dstVar=voice->tmp16_seq
         voice->tmp28_seq = voice->mod_mod_amp_cur;
         voice->tmp16_seq *= voice->tmp28_seq;
         voice->tmp14_rate += voice->tmp16_seq;
         
         // -- mod="$v_carrier_kbd" dstVar=voice->tmp14_rate
         voice->tmp29_seq = voice->var_v_carrier_kbd;
         voice->tmp14_rate *= voice->tmp29_seq;
         voice->tmp10_zlp_pos += int(voice->tmp14_rate * 65536);
   
         /* end calc */

#ifdef OVERSAMPLE_FACTOR
         outOS += out;
#endif // OVERSAMPLE_FACTOR
      }
#ifdef OVERSAMPLE_FACTOR
      // Apply lowpass filter before downsampling
      //   (note) normalized Fc = F/Fs = 0.442947 / sqrt(oversample_factor - 1)
      out = outOS * (1.0f / OVERSAMPLE_FACTOR);
#endif // OVERSAMPLE_FACTOR
      _samplesOut[k]      = out;
      _samplesOut[k + 1u] = out;

      // Next frame
      k += 2u;
      voice->note_speed_cur += voice->note_speed_inc;
      voice->mod_carrier_wave_cur  += voice->mod_carrier_wave_inc;
      voice->mod_carrier_freq_cur  += voice->mod_carrier_freq_inc;
      voice->mod_mod_wave_cur      += voice->mod_mod_wave_inc;
      voice->mod_mod_freq_cur      += voice->mod_mod_freq_inc;
      voice->mod_mod_amp_cur       += voice->mod_mod_amp_inc;
      voice->mod_mod_flt_cur       += voice->mod_mod_flt_inc;
   } /* loop numFrames */
}

static st_plugin_shared_t *ST_PLUGIN_API loc_shared_new(st_plugin_info_t *_info) {
   aon_fm_v1_shared_t *ret = (aon_fm_v1_shared_t *)malloc(sizeof(aon_fm_v1_shared_t));
   if(NULL != ret)
   {
      memset((void*)ret, 0, sizeof(*ret));
      ret->base.info  = _info;
      memcpy((void*)ret->params, (void*)loc_param_resets, NUM_PARAMS * sizeof(float));
   }
   return &ret->base;
}

static void ST_PLUGIN_API loc_shared_delete(st_plugin_shared_t *_shared) {
   free((void*)_shared);
}

static st_plugin_voice_t *ST_PLUGIN_API loc_voice_new(st_plugin_info_t *_info) {
   aon_fm_v1_voice_t *voice = (aon_fm_v1_voice_t *)malloc(sizeof(aon_fm_v1_voice_t));
   if(NULL != voice)
   {
      memset((void*)voice, 0, sizeof(*voice));
      voice->base.info = _info;
      voice->tmp10_zlp_pos = 0;
      voice->tmp17_zlp_pos = 0;
   }
   return &voice->base;
}

static void ST_PLUGIN_API loc_voice_delete(st_plugin_voice_t *_voice) {
   ST_PLUGIN_VOICE_CAST(aon_fm_v1_voice_t);

   free((void*)_voice);
}

static void ST_PLUGIN_API loc_plugin_exit(st_plugin_info_t *_info) {
   free((void*)_info);
}

#ifdef USE_CYCLE_SINE_TBL
static void loc_calc_sine_tbl(void) {
#define QCOS_BITS 16
#define QCOS_ONE  (1 << QCOS_BITS)
#define QCOS_MASK (QCOS_ONE - 1)
   // (note) same as in TSR "C" implementation
   float *qcos = cycle_sine_tbl_f + 4096; // quarter cos tbl

   // calc quarter cos tbl
   int k = 0;
   for(int a = 0; a < (QCOS_ONE/4); a += (QCOS_ONE/16384)/*4*/)  // 4096 elements
   {
      int x = a - ((int)(QCOS_ONE * 0.25 + 0.5)) + ((a + ((int)(QCOS_ONE * 0.25 + 0.5)))&~QCOS_MASK);
      x = ((x * ((x<0?-x:x) - ((int)(QCOS_ONE * 0.5 + 0.5)))) >> QCOS_BITS) << 4;
      x += (((((int)(QCOS_ONE * 0.225 + 0.5)) * x) >> QCOS_BITS) * ((x<0?-x:x) - QCOS_ONE)) >> QCOS_BITS;
      qcos[k++] = x / float(QCOS_ONE);
   }

   // 0..90 (rev qtbl)
   int j = 4096;
   k = 0;
   loop(4096)
      cycle_sine_tbl_f[k++] = qcos[--j];

   // 90..180 = cos tbl
   k += 4096;

   // 180..360 (y-flip first half of tbl)
   j = 0;
   loop(8192)
      cycle_sine_tbl_f[k++] = -cycle_sine_tbl_f[j++];

#ifdef USE_CYCLE_SINE_TBL_I
   loop(16384)
      cycle_sine_tbl_i[i] = (short)(cycle_sine_tbl_f[i] * 2048.0f/*FX_ONE*/);
#endif // USE_CYCLE_SINE_TBL_I
}
#endif // USE_CYCLE_SINE_TBL

extern "C" {
st_plugin_info_t *aon_fm_v1_init(void) {
   aon_fm_v1_info_t *ret = (aon_fm_v1_info_t *)malloc(sizeof(aon_fm_v1_info_t));

   if(NULL != ret)
   {
      memset(ret, 0, sizeof(*ret));

      ret->base.api_version = ST_PLUGIN_API_VERSION;
      ret->base.id          = "aon_fm_v1";  // unique id. don't change this in future builds.
      ret->base.author      = "bsp";
      ret->base.name        = "aon_fm_v1";
      ret->base.short_name  = "aon_fm_v1";
      ret->base.flags       = ST_PLUGIN_FLAG_OSC;
      ret->base.category    = ST_PLUGIN_CAT_UNKNOWN;
      ret->base.num_params  = NUM_PARAMS;
      ret->base.num_mods    = NUM_MODS;

      ret->base.shared_new                        = &loc_shared_new;
      ret->base.shared_delete                     = &loc_shared_delete;
      ret->base.voice_new                         = &loc_voice_new;
      ret->base.voice_delete                      = &loc_voice_delete;
      ret->base.get_param_name                    = &loc_get_param_name;
      ret->base.get_param_group_name              = &loc_get_param_group_name;
      ret->base.get_param_group_idx               = &loc_get_param_group_idx;
      ret->base.get_param_reset                   = &loc_get_param_reset;
      ret->base.query_dynamic_param_preset_values = &loc_query_dynamic_param_preset_values;
      ret->base.query_dynamic_param_preset_name   = &loc_query_dynamic_param_preset_name;
      ret->base.get_param_value                   = &loc_get_param_value;
      ret->base.set_param_value                   = &loc_set_param_value;
      ret->base.get_mod_name                      = &loc_get_mod_name;
      ret->base.set_sample_rate                   = &loc_set_sample_rate;
      ret->base.note_on                           = &loc_note_on;
      ret->base.set_mod_value                     = &loc_set_mod_value;
      ret->base.prepare_block                     = &loc_prepare_block;
      ret->base.process_replace                   = &loc_process_replace;
      ret->base.plugin_exit                       = &loc_plugin_exit;

#ifdef USE_CYCLE_SINE_TBL
      loc_calc_sine_tbl();
#endif // USE_CYCLE_SINE_TBL
   }

   return &ret->base;
}

#ifndef STFX_SKIP_MAIN_INIT
ST_PLUGIN_APICALL st_plugin_info_t *ST_PLUGIN_API st_plugin_init(unsigned int _pluginIdx) {
   switch(_pluginIdx)
   {
      case 0u:
         return aon_fm_v1_init();
   }
   return NULL;
}
#endif // STFX_SKIP_MAIN_INIT

} // extern "C"
